# SPDX-FileCopyrightText: 2024 Alec Delaney
# SPDX-License-Identifier: MIT

"""Configuration settings for sphinx."""

import sphinx_rtd_theme

project = "cronberry"
copyright = "2024 Alec Delaney"
author = "Alec Delaney"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
    # "autoapi.extension",
]

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for Sphinx AutoAPI------------------------------------------------
# autoapi_dirs = ["../cronberry"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path(), "_static"]
# html_static_path = ["_static"]



# SPDX-FileCopyrightText: 2024 Alec Delaney
# SPDX-License-Identifier: MIT

"""Pytest configuration file."""

import pathlib
import subprocess
from typing import Union

import pytest

BACKUP_FILEPATH = "tests/backup/table_backup.tab"


def pytest_sessionstart(session: pytest.Session) -> None:
    """Save the current cron table before testing."""
    proc = subprocess.Popen(
        ["crontab", "-l"], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
    )
    output, _ = proc.communicate()

    if output:  # pragma: no cover
        backup = pathlib.Path(BACKUP_FILEPATH)
        backup.parent.mkdir(parents=True, exist_ok=True)
        backup.write_bytes(output)

        proc = subprocess.Popen(
            ["crontab", "-r"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        proc.communicate()
        if proc.returncode:
            raise OSError("Error occurred while deleting the current crontab")


def pytest_sessionfinish(
    session: pytest.Session, exitstatus: Union[int, pytest.ExitCode]
) -> None:
    """Restore the previous cron table after testing."""
    proc = subprocess.Popen(
        ["crontab", "-r"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )
    proc.communicate()

    backup = pathlib.Path(BACKUP_FILEPATH)
    if backup.exists():  # pragma: no cover
        proc = subprocess.Popen(
            ["crontab", str(backup.resolve())],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        proc.communicate()
        if proc.returncode:
            raise OSError("Error occurred while restoring the previous crontab")
        backup.unlink()



# SPDX-FileCopyrightText: 2024 Alec Delaney
# SPDX-License-Identifier: MIT

"""Library for working with multiple cron jobs within a single cron table."""

import functools
import os
import re
import subprocess
import tempfile
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from typing_extensions import TypeAlias

from . import fields

CRONJOB_REGEX = r"^# \[(.*)]\n(.*)$"


Timing: TypeAlias = Union[fields.ShorthandSyntax, fields.ExplicitTiming]
"""Type alias for any kind of timing specification."""

Command: TypeAlias = str
"""Type alias for the cron job command."""


class CronJob:
    """Cron job representation."""

    def __init__(self, title: str, timing: Timing, command: Command):
        """Initialize the cron job object."""
        self.title = title
        self.timing = timing
        self.command = command

    def reduce(self) -> None:
        """Reduce duplicate timing instructions."""
        raise NotImplementedError("This feature is not yet implemented.")

    def __str__(self) -> str:
        """Representation of the cron job as a string (cron job)."""
        return f"{self.timing} {self.command}"

    def __repr__(self) -> str:
        """Printable representation of the cron job."""
        return f"CronJob(title={self.title!r}, timing={self.timing}, command={self.command!r})"

    def __eq__(self, __value: object) -> bool:
        """Check equality."""
        if isinstance(__value, CronJob):
            if (
                self.title == __value.title
                and self.timing == __value.timing
                and self.command == __value.command
            ):
                return True
        return False

    @staticmethod
    def parse_cron_text(cron_text: str) -> Tuple[Timing, Command]:
        """Parse a cron job in textual format."""
        try:
            shorthand = cron_text.split(" ")[0]
            timing = fields.ShorthandSyntax(shorthand)
            remaining_cron_texts = cron_text.split()[1:]
            remaining_cron_text = " ".join(remaining_cron_texts)
        except ValueError:
            timing_list = []
            remaining_cron_text = cron_text
            for _ in range(5):
                comp, _, remaining_cron_text = remaining_cron_text.partition(" ")
                parsed_comp = CronJob._parse_discrete_value(comp)
                timing_list.append(parsed_comp)
            timing = fields.ExplicitTiming(*timing_list)
        return timing, remaining_cron_text

    @classmethod
    def from_job(cls, job_text: str, title: str) -> "CronJob":
        """Create a CronJob from a cron job in texual format with a provided title."""
        timing, command = cls.parse_cron_text(job_text)
        return cls(title, timing, command)

    @staticmethod
    def _parse_discrete_value(value: str) -> List[fields.DiscreteValue]:
        """Parse the timing field values, recursively if needed."""
        values = []
        list_values = value.split(",")
        for list_value in list_values:
            if "/" in list_value:
                duration, frequency = list_value.split("/")
                dur_comp = CronJob._parse_discrete_value(duration)[0]
                freq_comp = CronJob._parse_discrete_value(frequency)[0]
                values.append(fields.StepValues(dur_comp, freq_comp))
            elif "-" in list_value:
                start, end = list_value.split("-")
                start_comp = CronJob._parse_discrete_value(start)[0]
                end_comp = CronJob._parse_discrete_value(end)[0]
                values.append(fields.ValueRange(start_comp, end_comp))
            elif list_value == "*":
                values.append(None)
            else:
                try:
                    values.append(int(list_value))
                except ValueError as err:
                    raise ValueError(
                        f"Could not parse timing component: {list_value}"
                    ) from err

        return values

    def to_job(self) -> Tuple[str, str]:
        """Get the job syntax, with the accompanying title."""
        if isinstance(self.timing, fields.ShorthandSyntax):
            timing_text = self.timing.value
        else:
            timing_text = str(self.timing)
        return timing_text, self.command


def parse_crontab(filepath: Optional[str] = None) -> List[CronJob]:
    """Parse the user's crontab."""
    if filepath is None:
        proc = subprocess.Popen(
            ["crontab", "-l"], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
        )
        output, _ = proc.communicate()
        text = output.decode()
    else:
        with open(filepath, encoding="utf-8") as jobfile:
            text = jobfile.read()

    re_matches: List[Tuple[str, str]] = re.findall(CRONJOB_REGEX, text, re.M)
    if not re_matches:
        return tuple()

    jobs = []
    for title, cron_text in re_matches:
        job = CronJob.from_job(cron_text, title)
        jobs.append(job)
    return jobs


def add_cronjobs(
    jobs: Iterable[CronJob], filepath: Optional[str] = None, *, overwrite: bool = False
) -> None:
    """Add the cron jobs to a crontab."""
    dest_crontab = parse_crontab(filepath)
    dest_titles = {dest_job.title for dest_job in dest_crontab}
    src_titles = {src_job.title for src_job in jobs}

    if len(src_titles) != len(jobs):
        raise ValueError("Cron jobs to be written contain duplicate titles")

    if len(dest_titles) != len(dest_crontab):
        raise RuntimeError(
            "Destination crontab already contains multiple entries, please remove duplicate"
        )

    has_overlap = dest_titles != dest_titles.difference(src_titles)
    if has_overlap and not overwrite:
        raise ValueError(
            "Destination crontab already contains title(s) provided, please remove duplicates"
        )

    writeable_jobs = {job.title: job for job in dest_crontab}
    add_jobs = {job.title: job for job in jobs}
    writeable_jobs.update(add_jobs)

    _update_crontab(writeable_jobs, filepath)


def remove_cronjob(
    job_name: str,
    filepath: Optional[str] = None,
    *,
    ignore_missing: bool = False,
) -> None:
    """Remove the cronjobs from a crontab."""
    dest_crontab = parse_crontab(filepath)
    current_jobs = {job.title: job for job in dest_crontab}

    if job_name not in current_jobs and not ignore_missing:
        raise ValueError(
            f"Cron job '{job_name}' does not exist in the destination crontab"
        )
    elif job_name in current_jobs:
        del current_jobs[job_name]

    _update_crontab(current_jobs, filepath)


def _update_crontab(jobs: Dict[str, CronJob], filepath: Optional[str] = None) -> None:
    """Update the crontab file with the new contents."""
    tabtext = ""
    for index, job in enumerate(jobs.values()):
        if index != 0:
            tabtext += "\n"
        cron_timing, cron_cmd = job.to_job()
        tabtext += f"# [{job.title}]\n{cron_timing} {cron_cmd}\n"

    write_file = (
        functools.partial(tempfile.NamedTemporaryFile, delete=False)
        if filepath is None
        else functools.partial(open, filepath)
    )

    with write_file(mode="w", encoding="utf-8") as destfile:
        output_filepath = destfile.name
        destfile.write(tabtext)

    if filepath is None:
        proc = subprocess.Popen(
            ["crontab", output_filepath],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        proc.communicate()
        os.remove(output_filepath)


def clear_cronjobs(filepath: Optional[str] = None):
    """Clear all of the cronjobs in a crontab."""
    _update_crontab({}, filepath)


def write_cronjobs(jobs: Iterable[CronJob], filepath: Optional[str] = None) -> None:
    """(Over)write a specific crontab with provided jobs."""
    job_dict = {job.title: job for job in jobs}
    _update_crontab(job_dict, filepath)


def save_crontab(dest_filepath: str) -> None:
    """Save the current user's crontab to a file."""
    proc = subprocess.Popen(
        ["crontab", "-l"], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
    )
    output, _ = proc.communicate()
    with open(dest_filepath, mode="wb") as outfile:
        outfile.write(output)



# SPDX-FileCopyrightText: 2024 Alec Delaney
# SPDX-License-Identifier: MIT

"""CLI for working with cronberry library."""

from typing import Optional, Tuple

import click

import cronberry


@click.group()
def cli() -> None:
    """Entry point for the CLI."""


@cli.command()
@click.argument(
    "crontab",
    type=click.Path(dir_okay=False, resolve_path=True),
)
@click.option(
    "-f",
    "--filepath",
    type=click.Path(dir_okay=False, resolve_path=True),
    default=None,
    help="Filepath to use instead of the user's crontab",
)
@click.option(
    "-o",
    "--overwrite/--no-overwrite",
    default=False,
    help="Overwrite if the job currently exists in the crontab",
)
@click.option(
    "-t", "--title", default="", help="Title of specific job to add from CRONTAB"
)
def add(crontab: str, filepath: Optional[str], overwrite: bool, title: str) -> None:
    """Add a job from CRONTAB to the crontab being used."""
    jobs = cronberry.parse_crontab(crontab)
    if title:
        jobs = [job for job in jobs if job.title == title]
    cronberry.add_cronjobs(jobs, filepath, overwrite=overwrite)


@cli.command()
@click.argument("job_title")
@click.argument("cronjob")
@click.option(
    "-f",
    "--filepath",
    type=click.Path(dir_okay=False, resolve_path=True),
    default=None,
    help="Filepath to use instead of the user's crontab",
)
@click.option(
    "-o",
    "--overwrite/--no-overwrite",
    default=False,
    help="Overwrite if the job currently exists in the crontab",
)
def enter(
    job_title: str, cronjob: str, filepath: Optional[str], overwrite: bool
) -> None:
    """Add CRONJOB with the name JOB_TITLE to the crontab."""
    timing, command = cronberry.CronJob.parse_cron_text(cronjob)
    job = cronberry.CronJob(job_title, timing, command)
    cronberry.add_cronjobs((job,), filepath, overwrite=overwrite)


@cli.command()
@click.argument(
    "job_titles",
    nargs=-1,
)
@click.option(
    "-f",
    "--filepath",
    type=click.Path(dir_okay=False, resolve_path=True),
    default=None,
    help="Filepath to use instead of the user's crontab",
)
@click.option(
    "-i",
    "--ignore-missing/--no-ignore-missing",
    default=False,
    help="Ignore jobs missing from the crontab",
)
def remove(
    job_titles: Tuple[str], filepath: Optional[str], ignore_missing: bool
) -> None:
    """Remove JOB_TITLES from the crontab being used."""
    unique_titles = set(job_titles)
    if len(unique_titles) != len(job_titles):
        raise click.ClickException("Duplicate job titles provided")
    for job_title in job_titles:
        try:
            cronberry.remove_cronjob(job_title, filepath, ignore_missing=ignore_missing)
        except ValueError:
            raise click.ClickException(
                f"Job title {job_title} does not exist in the crontab"
            )


@cli.command()
@click.argument(
    "filepath",
    type=click.Path(dir_okay=False, writable=True, resolve_path=True),
)
def save(filepath: str) -> None:
    """Save the current crontab being used to FILEPATH."""
    cronberry.save_crontab(filepath)


@cli.command()
@click.option(
    "-f",
    "--filepath",
    type=click.Path(dir_okay=False, writable=True, resolve_path=True),
    default=None,
    help="Filepath to use instead of the user's crontab",
)
def clear(filepath: Optional[str]) -> None:
    """Clear the current crontab being used."""
    cronberry.clear_cronjobs(filepath)


@cli.command()
@click.argument("job_title")
@click.option(
    "-f",
    "--filepath",
    type=click.Path(dir_okay=False, resolve_path=True),
    default=None,
    help="Filepath to use instead of the user's crontab",
)
def job(job_title: str, filepath: Optional[str]) -> None:
    """Get the cronbjob JOB_TITLE from a crontab."""
    all_jobs = cronberry.parse_crontab(filepath)
    selected_jobs = [any_job for any_job in all_jobs if any_job.title == job_title]
    if len(selected_jobs) < 1:
        raise click.ClickException(
            f"Job title {job_title} was not present in the crontab"
        )
    elif len(selected_jobs) > 1:
        raise click.ClickException(
            f"Job title {job_title} was found multiple times in the crontab"
        )
    selected_job = selected_jobs[0]
    formatted_command = f"{selected_job.timing!s} {selected_job.command!s}"
    click.echo(formatted_command)


@cli.command()
@click.option(
    "-f",
    "--filepath",
    type=click.Path(dir_okay=False, resolve_path=True),
    default=None,
    help="Filepath to use instead of the user's crontab",
)
def jobs(filepath: Optional[str]) -> None:
    """Get the jobs from a specific crontab."""
    jobs = cronberry.parse_crontab(filepath)
    for job in jobs:
        click.echo(job.title)



# SPDX-FileCopyrightText: 2024 Alec Delaney
# SPDX-License-Identifier: MIT

"""Tests for functionality related to cli functionality."""

import os
import subprocess
from typing import Tuple

from click.testing import CliRunner

from cronberry.app import cli


def crontab_remove() -> int:
    """Use crontab remove command."""
    proc = subprocess.Popen(
        ["crontab", "-r"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    proc.communicate()
    return proc.returncode


def crontab_list() -> Tuple[int, bytes, bytes]:
    """Use crontab list command."""
    proc = subprocess.Popen(
        ["crontab", "-l"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    output, error = proc.communicate()
    return proc.returncode, output, error


def crontab_set(filepath: str) -> int:
    """Use crontab list command."""
    proc = subprocess.Popen(
        ["crontab", filepath],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    proc.communicate()
    return proc.returncode


def test_cronberry_clear() -> None:
    """Tests the clear command."""
    src_filepath = "tests/tables/src.tab"
    runner = CliRunner()

    crontab_set(src_filepath)

    result = runner.invoke(cli, ["clear"])
    assert result.exit_code == 0

    _, output, _ = crontab_list()
    assert len(output) == 0


def test_cronberry_save() -> None:
    """Tests the save command."""
    src_filepath = "tests/tables/src.tab"
    dest_filepath = "tests/tables/temp.tab"
    runner = CliRunner()

    crontab_set(src_filepath)

    result = runner.invoke(cli, ["save", dest_filepath])
    assert result.exit_code == 0

    with open(src_filepath, "rb") as srcfile, open(
        dest_filepath, mode="rb"
    ) as destfile:
        src = srcfile.read()
        dest = destfile.read()
    assert src == dest

    os.remove(dest_filepath)


def test_cronberry_add() -> None:
    """Tests the add command."""
    src_filepath = "tests/tables/src.tab"
    new_filepath = "tests/tables/new.tab"
    comb_filepath = "tests/tables/combined.tab"
    runner = CliRunner()

    crontab_remove()

    result = runner.invoke(cli, ["add", src_filepath])
    assert result.exit_code == 0

    _, output, _ = crontab_list()
    with open(src_filepath, mode="rb") as srcfile:
        src_contents = srcfile.read()
    assert output == src_contents

    result = runner.invoke(cli, ["add", new_filepath])
    assert result.exit_code == 0

    _, output, _ = crontab_list()
    with open(comb_filepath, mode="rb") as combfile:
        comb_contents = combfile.read()
    assert output == comb_contents

    crontab_remove()

    # Test adding specific jobs
    crontab_set(new_filepath)
    result = runner.invoke(cli, ["add", new_filepath, "-t", "Job 5"])
    assert result.exit_code == 0

    result = runner.invoke(cli, ["add", new_filepath, "-t", "Job 6"])
    assert result.exit_code == 0

    _, output, _ = crontab_list()
    with open(new_filepath, mode="rb") as newfile:
        new_contents = newfile.read()
    assert output == new_contents

    crontab_remove()


def test_cronberry_remove() -> None:
    """Tests removing remove command."""
    src_filepath = "tests/tables/src.tab"
    comb_filepath = "tests/tables/combined.tab"
    runner = CliRunner()

    crontab_set(comb_filepath)

    result = runner.invoke(cli, ["remove", "Test 5", "Test 6"])
    assert result.exit_code == 0

    _, output, _ = crontab_list()

    with open(src_filepath, mode="rb") as srcfile:
        src_contents = srcfile.read()

    assert src_contents == output

    # Test errors

    result = runner.invoke(cli, ["remove", "Duplicate", "Duplicate"])
    assert result.exit_code != 0

    result = runner.invoke(cli, ["remove", "Not in table"])
    assert result.exit_code != 0

    crontab_remove()


def test_cronberry_job() -> None:
    """Tests the job command."""
    src_filepath = "tests/tables/src.tab"
    bad_filepath = "tests/tables/bad.tab"
    expected_job = '3-45/3 * * */2 * echo "Test 3!"\n'
    runner = CliRunner()

    crontab_set(src_filepath)

    result = runner.invoke(cli, ["job", "Test 3"])
    assert result.exit_code == 0
    assert result.output == expected_job

    # Test errors

    result = runner.invoke(cli, ["job", "Not in table"])
    assert result.exit_code != 0

    crontab_set(bad_filepath)
    result = runner.invoke(cli, ["job", "Test 7"])
    assert result.exit_code != 0

    crontab_remove()


def test_cronberry_jobs() -> None:
    """Tests the jobs command."""
    src_filepath = "tests/tables/src.tab"
    expected_jobs = "Test 1\nTest 2\nTest 3\nTest 4\n"
    runner = CliRunner()

    crontab_set(src_filepath)

    result = runner.invoke(cli, ["jobs"])
    assert result.exit_code == 0
    assert result.output == expected_jobs

    crontab_remove()


def test_cronberry_enter() -> None:
    """Tests the enter command."""
    job_title = "Manual"
    command = "1 2 3 4 5 echo Test"
    runner = CliRunner()

    result = runner.invoke(cli, ["enter", job_title, command])
    assert result.exit_code == 0

    _, output, _ = crontab_list()
    assert output.decode() == f"# [{job_title}]\n{command}\n"

    crontab_remove()



# SPDX-FileCopyrightText: 2024 Alec Delaney
# SPDX-License-Identifier: MIT

"""Tests for functionality related to library functionality."""

import copy
import os
import tempfile
from typing import List

import pytest

import cronberry


def test_read_crontab() -> None:
    """Tests whether crontabs can be read correctly."""
    jobs = cronberry.parse_crontab("tests/tables/new.tab")

    first_job = jobs[0]

    assert first_job.title == "Test 5"
    assert first_job.command == 'echo "Test 5!"'

    timing = first_job.timing
    assert isinstance(timing, cronberry.fields.ExplicitTiming)
    assert timing.minute == [8]
    assert timing.hour == [8]
    assert timing.day_of_month == [3]
    assert timing.month == [7]
    assert timing.day_of_week == [2]

    second_job = jobs[1]

    assert second_job.title == "Test 6"
    assert second_job.command == 'echo "Test 6!"'

    timing = second_job.timing
    assert isinstance(timing, cronberry.fields.ExplicitTiming)
    assert timing.minute == [1]
    assert timing.hour == [1]
    assert timing.day_of_month == [1]
    assert timing.month == [None]
    assert timing.day_of_week == [None]


def test_cronjob_equality() -> None:
    """Test the equality of two CronJobs."""
    jobs = cronberry.parse_crontab("tests/tables/src.tab")

    original_job = jobs[0]
    copy_job = copy.deepcopy(original_job)

    assert copy_job == original_job

    temp_timing: List[cronberry.fields.DiscreteValue] = list(copy_job.timing)
    temp_timing[3] = [7, 8]
    altered_timing = cronberry.fields.ExplicitTiming(*temp_timing)
    assert altered_timing != copy_job.timing

    copy_job.timing = altered_timing
    assert copy_job != original_job

    assert altered_timing.__eq__(42) is False
    assert copy_job.__eq__(42) is False
    assert cronberry.fields.ValueRange(1, 2).__eq__(42) is False
    assert cronberry.fields.StepValues(1, 2).__eq__(42) is False


def test_cronjob_str() -> None:
    """Test the ability to convert a CronJob to its string equivalent."""
    jobs = cronberry.parse_crontab("tests/tables/src.tab")
    job = jobs[0]

    with open("tests/tables/src.tab", encoding="utf-8") as textfile:
        cronjob_text = textfile.read().split("\n")[1]
    assert str(job) == cronjob_text.strip()


def test_bad_parse() -> None:
    """Tests that an error is raised when a bad values is parsed."""
    with pytest.raises(ValueError):
        cronberry.CronJob._parse_discrete_value("#42")


def test_cronjob_reduce() -> None:
    """Tests that the CronJob reduce functionality works."""
    with pytest.raises(NotImplementedError):
        job = cronberry.CronJob(
            "Test", cronberry.fields.ShorthandSyntax.DAILY, 'echo "Test!'
        )
        job.reduce()


def test_write_crontab() -> None:
    """Tests the ability to (over)write a crontab."""
    overwrite_jobs = cronberry.parse_crontab("tests/tables/new.tab")
    cronberry.write_cronjobs(overwrite_jobs)
    read_jobs = cronberry.parse_crontab()
    assert read_jobs == overwrite_jobs


def test_clear_cronjobs() -> None:
    """Tests the ability to clear cronjobs from a crontab."""
    overwrite_jobs = cronberry.parse_crontab("tests/tables/new.tab")
    cronberry.write_cronjobs(overwrite_jobs)
    cronberry.clear_cronjobs()
    assert not cronberry.parse_crontab()


def test_writeread_equality() -> None:
    """Tests that writing a cronjob then reading it is lossless."""
    file_jobs = cronberry.parse_crontab("tests/tables/new.tab")
    cronberry.write_cronjobs(file_jobs)
    nonfile_jobs = cronberry.parse_crontab()
    assert nonfile_jobs == file_jobs
    cronberry.clear_cronjobs()


def test_remove_cronjob() -> None:
    """Tests removing a cronjob from a file."""
    file_jobs = cronberry.parse_crontab("tests/tables/src.tab")
    cronberry.write_cronjobs(file_jobs)
    cronberry.remove_cronjob("Test 2")
    remaining_jobs = cronberry.parse_crontab()
    remaining_jobs_dict = {job.title: job for job in remaining_jobs}
    for index, (key, exp) in enumerate(
        (
            ("Test 1", True),
            ("Test 2", False),
            ("Test 3", True),
            ("Test 4", True),
        )
    ):
        assert (key in remaining_jobs_dict) == exp
        if exp:
            assert remaining_jobs_dict[key] == file_jobs[index]

    with pytest.raises(ValueError):
        cronberry.remove_cronjob("Test 2")

    cronberry.clear_cronjobs()


def test_add_cronjob() -> None:
    """Tests adding a cronjob from a file."""
    file_jobs = cronberry.parse_crontab("tests/tables/src.tab")
    add_file_jobs = cronberry.parse_crontab("tests/tables/new.tab")
    cronberry.write_cronjobs(file_jobs)

    combined_jobs = file_jobs + add_file_jobs
    cronberry.add_cronjobs(add_file_jobs)

    assert cronberry.parse_crontab() == combined_jobs

    # Test errors

    with pytest.raises(RuntimeError):
        cronberry.add_cronjobs(add_file_jobs, "tests/tables/bad.tab")

    add_file_jobs[0].command = "echo \"This wasn't here before!"

    with pytest.raises(ValueError):
        cronberry.add_cronjobs(add_file_jobs)

    add_file_jobs[0].title = "Test 6"
    cronberry.clear_cronjobs()

    with pytest.raises(ValueError):
        cronberry.add_cronjobs(add_file_jobs)

    cronberry.clear_cronjobs()


def test_save_crontab() -> None:
    """Tests the ability to save a crontab file."""
    file_jobs = cronberry.parse_crontab("tests/tables/src.tab")
    cronberry.write_cronjobs(file_jobs)
    temp_file = tempfile.NamedTemporaryFile(mode="r+", encoding="utf-8", delete=False)
    cronberry.save_crontab(temp_file.name)
    temp_file.seek(0)

    with open("tests/tables/src.tab", mode="r+") as origfile:
        orig_contents = origfile.read()
        save_contents = temp_file.read()

    assert save_contents == orig_contents

    os.remove(temp_file.name)


def test_cronjob_repr() -> None:
    """Tests CrobJob.__repr__()."""
    EXPECTED_REPR = "CronJob(title='Test 4', timing=ShorthandSyntax.YEARLY, command='echo \"Test 3!\"')"
    target_job = cronberry.parse_crontab("tests/tables/src.tab")[-1]
    assert repr(target_job) == EXPECTED_REPR



# SPDX-FileCopyrightText: 2024 Alec Delaney
# SPDX-License-Identifier: MIT

"""Classes and type annotations for cronjob fields."""

from enum import Enum
from typing import Any, NamedTuple, Tuple, Union

from typing_extensions import TypeAlias


def _cron_str(comp: "DiscreteValue") -> str:
    """Return the cron equivalent string of a discrete value (field value)."""
    return "*" if comp is None else str(comp)


class ShorthandSyntax(Enum):
    """Enum for shorthand syntax of timing options."""

    HOURLY = "@hourly"
    DAILY = "@daily"
    WEEKLY = "@weekly"
    ANNUALLY = "@annually"
    YEARLY = "@yearly"
    REBOOT = "@reboot"


AnyValue: TypeAlias = None
"""Type alias for any value (*)."""

ExactValue: TypeAlias = int
"""Type alias for an exact value (2)."""


class ValueRange(NamedTuple):
    """Named tuple of a range of values (3-5)."""

    start: ExactValue
    end: ExactValue

    def __eq__(self, other: Any) -> bool:
        """Check equality."""
        if isinstance(other, ValueRange):
            return self.start == other.start and self.end == other.end
        return False

    def __str__(self) -> str:
        """Representation of the value range as a string (x-y)."""
        return f"{self.start}-{self.end}"


class StepValues(NamedTuple):
    """Named tuple of a set of step values (8/3)."""

    duration: Union[AnyValue, ExactValue, ValueRange]
    frequency: ExactValue

    def __eq__(self, other: Any) -> bool:
        """Check equality."""
        if isinstance(other, StepValues):
            return self.duration == other.duration and self.frequency == other.frequency
        return False

    def __str__(self) -> str:
        """Representation of the step values as a string (x/y)."""
        return f"{_cron_str(self.duration)}/{self.frequency}"


ListValue: TypeAlias = Union[AnyValue, ExactValue, ValueRange, StepValues]
"""Type alias of any single list item."""

ListValues: TypeAlias = Tuple[ListValue, ...]
"""Type alias for a list of items."""

DiscreteValue: TypeAlias = Union[ListValues, ListValue]
"""Type alias for a single discrete value (field value)."""


class ExplicitTiming(NamedTuple):
    """Named tuple for an explicit timing setup."""

    minute: DiscreteValue
    hour: DiscreteValue
    day_of_month: DiscreteValue
    month: DiscreteValue
    day_of_week: DiscreteValue

    def __eq__(self, __value: object) -> bool:
        """Check equality."""
        if isinstance(__value, ExplicitTiming):
            return (
                self.minute == __value.minute
                and self.hour == __value.hour
                and self.day_of_month == __value.day_of_month
                and self.month == __value.month
                and self.day_of_week == __value.day_of_week
            )
        return False

    def __str__(self) -> str:
        """Representation of the explicit timing as a string (a, b, c, d, e)."""
        minute_comp = ",".join(_cron_str(comp) for comp in self.minute)
        hour_comp = ",".join(_cron_str(comp) for comp in self.hour)
        dom_comp = ",".join(_cron_str(comp) for comp in self.day_of_month)
        month_comp = ",".join(_cron_str(comp) for comp in self.month)
        dow_comp = ",".join(_cron_str(comp) for comp in self.day_of_week)
        return f"{minute_comp} {hour_comp} {dom_comp} {month_comp} {dow_comp}"
